@@ -3,7 +3,6 @@ use ec::State;
 use regex::Regex;
 use std::cmp::min;
 use std::io;
-use std::io::Read;
 use std::io::Write;
 use std::num::NonZeroUsize;
 
@@ -35,47 +34,74 @@ struct Command {
 
 fn print_help() {
     print!("Input/output is hex unless toggled to decimal with 'x'
-h            This (h)elp
-<Enter>      Print current byte(s) and move forward to next set of byte(s)
-3d4          Move to byte number 3d4 and print from there
-+            Move 1 byte forward and print from there
-+++          Move 3 bytes forward and print from there
--            Move 1 byte back and print from there
-+3d4         Move 3d4 bytes forward and print from there
--3d4         Move 3d4 bytes back and print from there
-$            Move to last byte and print it
-/deadbeef    If the bytes de ad be ef exist after the current index, move there
-               and print.
-?deadbeef    If the bytes de ad be ef exist before the current index, move there
-               and print.
-/            Perform last search again starting at next byte
-?            Perform last search (backwards) again starting at previous byte
-k            Delete/(k)ill byte at current index and print new line of byte(s)
-7dk          Move to byte 7d, (k)ill that byte, and print from there.
-1d,72k       Move to byte 1d, (k)ill bytes 1d - 72 inclusive, and print from there.
-/deadbeef/k  If the bytes de ad be ef exist after the current index, move there,
-               (k)ill those bytes, and print.
-i            Prompt you to write out bytes which will be (i)nserted at current index
-72i          Move to byte number 72 and prompt you to enter bytes which will be
-               (i)nserted there.
-/deadbeef/i  If the bytes de ad be ef exist after the current index, move there
-               and prompt you to enter bytes which will be (i)nserted there.
-12,3dp       (p)rint bytes 12 - 3d inclusive, move to byte 12
-m            Toggle whether or not characters are printed after bytes
-n            Toggle whether or not byte (n)umbers are printed before bytes
-o            Toggle using c(o)lor
-p            (p)rint current line of byte(s) (depending on 'W')
-s            Print (s)tate of all toggles and 'W'idth
-t3d          Print 0x3d lines of con(t)extual bytes after current line [Default 0]
-T3d          Print 0x3d lines of con(T)extual bytes before current line [Default 0]
-x            Toggle interpreting inputs and displaying output as he(x) or decimal
-w            Actually (w)rite changes to the file on disk
-W3d          Set (W)idth to 0x3d.  i.e. print a linebreak every 3d bytes [Default 0x10]
-q            (q)uit
+h           This (h)elp
+<Enter>     Print current byte(s) and move forward to next line of byte(s)
+j           (j)ump back to previous line of byte(s) and print
+3d4         Move to byte number 3d4 and print from there
++           Move 1 byte forward and print from there
++++         Move 3 bytes forward and print from there
+-           Move 1 byte back and print from there
++3d4        Move 3d4 bytes forward and print from there
+-3d4        Move 3d4 bytes back and print from there
+$           Move to last byte and print it
+/deadbeef   If bytes de ad be ef exist after current index, move there and print
+?deadbeef   If bytes de ad be ef exist before current index, move there and print
+/           Perform last search again starting at next byte
+?           Perform last search (backwards) again starting at previous byte
+k           Delete/(k)ill byte at current index and print new line of byte(s)
+7dk         Move to byte 7d, (k)ill that byte, and print from there.
+1d,72k      Move to byte 1d; (k)ill bytes 1d - 72 inclusive; print from there
+/deadbeef/k If bytes de ad be ef exist after current index, move there,
+              (k)ill those bytes, and print
+i           Prompt you to enter bytes which will be (i)nserted at current index
+72i         Move to byte number 72; prompt you to enter bytes to (i)nsert there
+/deadbeef/i If bytes de ad be ef exist after current index, move there
+              and prompt you to enter bytes which will be (i)nserted there
+12,3dp      (p)rint bytes 12 - 3d inclusive, move to byte 12
+m           Toggle whether or not characters are printed after bytes
+n           Toggle whether or not byte (n)umbers are printed before bytes
+o           Toggle using c(o)lor
+p           (p)rint current line of byte(s) (depending on 'W')
+R           Toggle (R)ead-only mode
+s           Print (s)tate of toggles, 'W'idth, etc.
+t3d         Print 0x3d lines of con(t)extual bytes after current line [Default 0]
+T3d         Print 0x3d lines of con(T)extual bytes before current line [Default 0]
+u           (u)pdate filename to write to
+x           Toggle reading input and displaying output as he(x) or decimal
+w           Actually (w)rite changes to the file on disk
+W3d         Set (W)idth to 0x3d.  i.e. print a linebreak every 3d bytes [Default 0x10]
+q           (q)uit
 ");
 }
 
 
+fn read_string_from_user(prompt: Option<&str>) -> Result<String, String> {
+    print!("{}", if prompt.is_none() {
+            "> "
+        }
+        else {
+            prompt.unwrap()
+        }
+    );
+
+    io::stdout().flush().unwrap();
+    let result = get_input_or_die();
+
+    if result.is_ok() {
+        Ok(result.unwrap())
+    }
+
+    /* Consider EOF empty string */
+    else if result == Err(0) {
+        Ok("".to_owned())
+    }
+
+    else {
+        Err("Couldn't read input from user".to_owned())
+    }
+}
+
+
 fn read_bytes_from_user() -> Result<Vec<u8>, String> {
     print!("> ");
     io::stdout().flush().unwrap();
@@ -107,7 +133,7 @@ impl Command {
         let re_search_again = Regex::new(r"^ *(?P<direction>[/?]) *$").unwrap();
         let re_search_kill = Regex::new(r"^ */(?P<bytes>[0-9a-fA-F]+)/k *$").unwrap();
         let re_search_insert = Regex::new(r"^ */(?P<bytes>[0-9a-fA-F]+)/i *$").unwrap();
-        let re_single_char_command = Regex::new(r"^ *(?P<command>[hmnopsxqwik]).*$").unwrap();
+        let re_single_char_command = Regex::new(r"^ *(?P<command>[hijkmnopqRsuwx]).*$").unwrap();
         let re_range = Regex::new(r"^ *(?P<begin>[0-9a-fA-F.$]+) *, *(?P<end>[0-9a-fA-F.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_specified_index = Regex::new(r"^ *(?P<index>[0-9A-Fa-f.$]+) *(?P<the_rest>.*) *$").unwrap();
         let re_offset_index = Regex::new(r"^ *(?P<sign>[-+])(?P<offset>[0-9A-Fa-f]+) *(?P<the_rest>.*) *$").unwrap();
@@ -578,7 +604,7 @@ fn minuses(state:&mut State, num_minuses:usize) -> Result<usize, String> {
     }
     else {
         state.index -= num_minuses;
-        state.print_bytes_sans_context(state.range(), false);
+        state.print_bytes();
         Ok(state.index)
     }
 }
@@ -599,7 +625,7 @@ fn pluses(state:&mut State, num_pluses:usize) -> Result<usize, String> {
                 }
                 else {
                     state.index += num_pluses;
-                    state.print_bytes_sans_context(state.range(), false);
+                    state.print_bytes();
                     Ok(state.index)
                 }
             },
@@ -611,46 +637,66 @@ fn pluses(state:&mut State, num_pluses:usize) -> Result<usize, String> {
 }
 
 
-pub fn actual_runtime(filename:&str, quiet:bool, color:bool) -> i32 {
-    let file = match ec::filehandle(filename) {
-        Ok(Some(filehandle)) => {
-            Some(filehandle)
-        },
-        Ok(None) => None,
-        Err(error) => {
-            println!("Problem opening '{}' ({:?})", filename, error);
-            return 3;
-        }
-    };
+pub fn cargo_version() -> Result<String, String> {
+    if let Some(version) = option_env!("CARGO_PKG_VERSION") {
+        return Ok(String::from(version));
+    }
+    return Err("Version unknown (not compiled with cargo)".to_string());
+}
 
-    let original_num_bytes = match ec::num_bytes_or_die(&file) {
-        Ok(num_bytes) => {
-            num_bytes
-        },
-        Err(errcode) => {
-            return errcode;
-        }
-    };
 
-    /* Read all bytes into memory just like real ed */
-    // TODO A real hex editor needs to buffer
-    let mut all_bytes = Vec::new();
-    if file.is_some() {
-        match file.unwrap().read_to_end(&mut all_bytes) {
-            Err(_) => {
-                println!("Couldn't read {}", filename);
-                return 4;
-            },
-            Ok(num_bytes_read) => {
-                if num_bytes_read != original_num_bytes {
-                    println!("Only read {} of {} bytes of {}", num_bytes_read,
-                            original_num_bytes, filename);
-                    return 5;
-                }
-            }
+pub fn update_filename(state: &mut ec::State) {
+    let filename = read_string_from_user(Some("Enter new filename: "));
+    if filename.is_err() {
+        println!("? {:?}", filename);
+        return;
+    }
+    let filename = filename.unwrap();
+
+    state.filename = filename;
+}
+
+
+pub fn write_out(state: &mut ec::State) {
+    if state.readonly {
+        println!("? Read-only mode");
+        return;
+    }
+    
+    /* Early return if write unsuccessful */
+    if state.filename != "" {
+        let result = std::fs::write(&state.filename, &state.all_bytes);
+        if result.is_err() {
+            println!("? (Couldn't write to {})", state.filename);
+            return;
         }
     }
+    else {
+        let filename = read_string_from_user(Some("Enter filename: "));
+        if filename.is_err() {
+            println!("? {:?}", filename);
+            return;
+        }
+        let filename = filename.unwrap();
 
+        /* filename is a string */
+        let result = std::fs::write(&filename, &state.all_bytes);
+        if result.is_err() {
+            println!("? (Couldn't write to given filename '{}')", filename);
+            return;
+        }
+
+        state.filename = filename;
+        println!("Write successfull, changing filename to '{}'", state.filename);
+    }
+
+    state.unsaved_changes = false;
+}
+
+
+/// If `filename` is "", open an empty buffer
+pub fn actual_runtime(filename:&str, quiet:bool, color:bool, readonly:bool)
+        -> i32 {
 
     // TODO Below here should be a function called main_loop()
     let mut state = ec::State{
@@ -660,12 +706,24 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool) -> i32 {
         show_prompt: !quiet,
         color: color,
         show_chars: true,
-        unsaved_changes: false,
+        unsaved_changes: (filename == ""),
+        readonly: readonly,
         index: 0,
         before_context: 0,
         after_context: 0,
         width: NonZeroUsize::new(16).unwrap(),
-        all_bytes: all_bytes,
+        all_bytes: if filename == "" {
+                Vec::new()
+            }
+            else {
+                let maybe_all_bytes = ec::all_bytes_from_filename(filename);
+                if maybe_all_bytes.is_err() {
+                    println!("{:?}", maybe_all_bytes);
+                    return 1;
+                }
+                maybe_all_bytes.unwrap()
+            }
+        ,
         // TODO calculate based on longest possible index
         n_padding: "      ".to_owned(),
         last_search: None,
@@ -772,6 +830,21 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool) -> i32 {
                         print_help();
                     },
 
+                    /* User wants to go up a line */
+                    'j' => {
+                        if state.empty() {
+                            println!("? (Empty file)");
+                            continue;
+                        };
+
+                        let width = usize::from(state.width);
+                        let first_byte_to_show_index =
+                                state.index.saturating_sub(width);
+                        state.index = first_byte_to_show_index;
+                        state.print_bytes();
+                    }
+
+
                     /* 'k'ill byte(s) (Can't use 'd' because that's a hex
                     * character! */
                     'k' => {
@@ -784,7 +857,8 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool) -> i32 {
                         right_half = right_half.split_off(command.range.1 - command.range.0 + 1);
                         state.all_bytes.append(&mut right_half);
                         state.index = command.range.0;
-                        state.print_bytes_sans_context(state.range(), false);
+                        state.unsaved_changes = true;
+                        state.print_bytes();
                     },
 
                     /* Toggle showing char representations of bytes */
@@ -897,6 +971,12 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool) -> i32 {
                         return 0;
                     },
 
+                    /* Toggle readonly mode */
+                    'R' => {
+                        state.readonly = !state.readonly;
+                    },
+
+
                     /* Print state */
                     's' => {
                         println!("{}", state);
@@ -912,12 +992,14 @@ pub fn actual_runtime(filename:&str, quiet:bool, color:bool) -> i32 {
                         state.before_context = usize::from(command.range.0);
                     },
 
+                    /* (u)pdate iflename */
+                    'u' => {
+                        update_filename(&mut state);
+                    },
+
                     /* Write out */
                     'w' => {
-                        let result = std::fs::write(filename, &state.all_bytes);
-                        if result.is_err() {
-                            println!("? (Couldn't write to {})", state.filename);
-                        }
+                        write_out(&mut state);
                     },
 
                     /* Change width */
